//---------------------------------------------------------------------------
// TDcjMessageDlg Component
//
// Copyright  1998, Michael Trier.  All Rights Reserved.
//---------------------------------------------------------------------------
// This component and its associated property editor and component editor can be
// freely used and distributed, provided all source code files are not modified
// in any way and there is no charge for it other than nominal shipping and
// handling fees.  Please contact me directly if you desire to modify the terms
// of this agreement.
//---------------------------------------------------------------------------
// Please note:  This source code is provided as a teaching tool. If you
// decide that you want to use whole routines or modules directly in an
// application that you are creating, please give credit where it is due.
// Given the above, keep in mind that these applications are copyright
// property of Michael Trier.
//---------------------------------------------------------------------------
// Date last modified: 10-Oct-1998
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// TDcjMessageDlg v0.80
//---------------------------------------------------------------------------
// Revision History:
// 0.30:    * Initial release.  Property editor doesn't work.
// 0.45:    * Property editor fixed.  It now works.
//          * Added a component editor that allows editting of buttons as well
//            as previewing the dialog.
//          * Added the component image.
// 0.60:    * Fixed problem with TCollection object not streaming published
//            properties.  Had to create a TButtonContainer object.
// 0.65:    * Buttons sizes were growing each time the message dialog was
//            displayed.  Corrected.
//          * Minor bug fixes.
// 0.70:    * When button name is selected in the property editor, it now sets
//            focus to the caption field.
//          * Added default button templates to the New button.
//          * I added a Preview button to the Buttons property editor, but
//            removed it because I realized that a Preview button would need to
//            inherently understand the functionality of the component object
//            that owns it.  This would restrict future uses for the Buttons
//            editor.
// 0.75:    * Added a ButtonID property that gets returned to the ButtonID
//            property of the TDcjMessageDlg when the dialog is closed.  This
//            allows for custom button Identifiers.  Note that the modal result
//            0f a custom button should be set to a value that closes the
//            dialog (i.e. not mrNone).
//          * I also changed the preview so that it displays the return results.
//          * Added a HelpContext property.  The HelpFile that is  used is the
//            application helpfile.  I may change this in the future.
// 0.77:    * Mades some serious modifications so that the component compiles
//            under BCB1.  Mainly, the TCollection object doesn't hava an ID
//            property.  I used the Index property instead.  It's not as
//            reliable, but works just as well for what I use it for.
//          * Pulled more hair out over BCB1 problems.  For some crazy reason
//            the component editor will not link unless a constructor is
//            declared.
// 0.80:    * I added a seperate registration unit, but removed it because
//            BCB won't link without including the other units anyway.  You'll
//            get unresolved externals in the linkage.  A shortcut is to use
//            the #pragma link directive, but that requires that you have
//            up-to-date object files.  In addition, if you include the other
//            units under BCB1, it will complain that they are not exporting
//            the Register() function.  Is any of this easy?
//          * I changed the TDcjButtonsEditor to descend from TComponentEditor
//            instead of TDefaultEditor.  How I made that mistake, I'll never
//            know.
//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <memory>
#pragma hdrstop

#include "DcjMsgDg.h"
#ifndef DCJ_BCB_1
  #pragma resource "*.res"
  #pragma package(smart_init)
#endif
//---------------------------------------------------------------------
static inline TDcjMessageDlg *ValidCtrCheck()
{
  return new TDcjMessageDlg(NULL);
}
//---------------------------------------------------------------------------
// TDcjMessageForm Implementation
//---------------------------------------------------------------------------
void __fastcall TDcjMessageForm::HelpButtonClick(System::TObject* Sender)
{
  Application->HelpContext(HelpContext);
}
//---------------------------------------------------------------------------
void __fastcall TDcjMessageForm::ButtonClick(System::TObject* Sender)
{
  // store the ButtonID for the button that was clicked
	FButtonID = dynamic_cast<TButton *>(Sender)->Tag;
}
//---------------------------------------------------------------------------
// TDcjButtonContainer Implementation
//---------------------------------------------------------------------------
__fastcall TDcjButtonContainer::TDcjButtonContainer(TComponent *Owner)
  : TPersistent()
{
  // store the container's owner
  FOwner = Owner;
  // create the buttons collection
  FButtons = new TDcjButtons(this);
  // set default properties
  FHorizontal = DEFAULT_BTN_HORIZ;
  FVertical = DEFAULT_BTN_VERT;
  FSpacing = DEFAULT_BTN_SPACING;
}
//---------------------------------------------------------------------------
__fastcall TDcjButtonContainer::~TDcjButtonContainer(void)
{
  // free up memory
  if(FButtons) delete FButtons;
}
//---------------------------------------------------------------------------
void __fastcall TDcjButtonContainer::Assign(TPersistent *Source)
{
  // assign the source to the destination object
  if( Source->InheritsFrom(__classid(TDcjButtonContainer)) ) {
    // cast the TPersistant object to a TDcjButtonContainer object
    TDcjButtonContainer *source = dynamic_cast<TDcjButtonContainer *>(Source);
    // assign the properties
    FHorizontal = source->Horizontal;
    FVertical = source->Vertical;
    FSpacing = source->Spacing;
    FButtons->Assign(source->Buttons);
  }
}
//---------------------------------------------------------------------------
TPersistent* __fastcall TDcjButtonContainer::GetOwner(void)
{
  return FOwner;
}
//---------------------------------------------------------------------------
// TDcjButtons Implementation
//---------------------------------------------------------------------------
__fastcall TDcjButtons::TDcjButtons(TPersistent *Owner)
  : TCollection(__classid(TDcjButton))
{
  // store the collections owner
  FOwner = Owner;
}
//---------------------------------------------------------------------------
void __fastcall TDcjButtons::Assign(TPersistent* Source)
{
  // assign the properties from one TDcjButtons to another
  if( Source->InheritsFrom(__classid(TDcjButtons)) ) {
    // make sure the destination object is clear
    Clear();

    // cast the TPersistant object to a TDcjButtons object
    TDcjButtons *source = dynamic_cast<TDcjButtons *>(Source);
    // create a button to hold the button items
    TDcjButton *button = NULL;
    // add the TDcjButton items
    for( int i = 0; i < source->Count; i++ ) {
      // add the button
      button = Add();
      button->Assign(source->Items[i]);
    }
  }
}
//---------------------------------------------------------------------------
TDcjButton* __fastcall TDcjButtons::Add(void)
{
  // override the Add member function, casting the collection item to a
  // TDcjButton.
  return dynamic_cast<TDcjButton *>(TCollection::Add());
}
//---------------------------------------------------------------------------
TDcjButton* __fastcall TDcjButtons::Add(AnsiString sCaption,
    TModalResult tResult, int iWidth, int iHeight)
{
  // overloads the add member function, passing in construction values
  TDcjButton *button = Add();
  button->Caption = sCaption;
  button->ModalResult = tResult;
  button->Width = iWidth;
  button->Height = iHeight;
  return button;
}
//---------------------------------------------------------------------------
TDcjButton* __fastcall TDcjButtons::GetItem(int Index)
{
  // override the GetItem member function, casting the collection item to
  // a TDcjButton.
  return dynamic_cast<TDcjButton *>(TCollection::GetItem(Index));
}
//---------------------------------------------------------------------------
void __fastcall TDcjButtons::SetItem(int Index, TDcjButton* Value)
{
  TCollection::SetItem(Index, Value);
}
//---------------------------------------------------------------------------
TPersistent* __fastcall TDcjButtons::GetOwner(void)
{
  return FOwner;
}
//---------------------------------------------------------------------------
void __fastcall TDcjButtons::DeleteItem(int ItemIndex)
{
  // delete the item at the index specified
  // there isn't any delete member function for the TCollection class, so
  // we have to copy the object except for the item specified.
  TDcjButtons *buttons = new TDcjButtons(FOwner);
  try {
    // create a button to hold the button items
    TDcjButton *button = NULL;
    // add the TDcjButton items
    for( int i = 0; i < Count; i++ ) {
      // check if the button is the one we don't want
      if( Items[i]->Index != ItemIndex ) {
        // add the button
        button = buttons->Add();
        button->Assign(Items[i]);
      }
    }

    // assign the new collection back to the old one
    this->Assign(buttons);
  }
  catch(...) {
    // delete the new collection
    delete buttons;
    // rethrow the exception
    throw;
  }
  // delete the new collection
  delete buttons;
}
//---------------------------------------------------------------------------
void __fastcall TDcjButtons::DeleteItem(TDcjButton *Item)
{
  // delete the collection item Item from the collection
  DeleteItem(Item->Index);
}

//---------------------------------------------------------------------------
// TDcjButton Implementation
//---------------------------------------------------------------------------
__fastcall TDcjButton::TDcjButton(TCollection *Collection)
  : TCollectionItem(Collection)
{
  // store the owner collection
  FCollection = dynamic_cast<TDcjButtons *>(Collection);
  // create the Dcj Button
  FWidth = DEFAULT_BTN_WIDTH;
  FHeight = DEFAULT_BTN_HEIGHT;
  FModalResult = DEFAULT_BTN_MODALRESULT;
  FDefault = false;
  FCancel = false;
  FHelp = false;
  FButtonID = 0;
}
//---------------------------------------------------------------------------
void __fastcall TDcjButton::SetCollection(TDcjButtons* Value)
{
}
//---------------------------------------------------------------------------
TDcjButtons* __fastcall TDcjButton::GetCollection(void)
{
  // return the owner collection
  return FCollection;
}
//---------------------------------------------------------------------------
void __fastcall TDcjButton::SetButtonID(TButtonID Value)
{
  // make sure it is not already set
  if( FButtonID != Value )
    FButtonID = Value;
}
//---------------------------------------------------------------------------
void __fastcall TDcjButton::SetWidth(int Value)
{
  if( FWidth != Value )
    FWidth = Value;
}
//---------------------------------------------------------------------------
void __fastcall TDcjButton::SetHeight(int Value)
{
  if( FHeight != Value )
    FHeight = Value;
}
//---------------------------------------------------------------------------
void __fastcall TDcjButton::SetDefault(bool Value)
{
  // make sure it is not already set
  if( FDefault != Value ) {
    // if true, clear any other button set to the default button
    if( Value ) {
      for( int i = 0; i < Collection->Count; i++ )
        if( (Collection->Items[i])->Default )
          Collection->Items[i]->Default = false;
    }
    // set this button as the default button
    FDefault = Value;
  }
}
//---------------------------------------------------------------------------
void __fastcall TDcjButton::SetCancel(bool Value)
{
  // make sure it is not already set
  if( FCancel != Value ) {
    // if true, clear any other button set to the cancel button
    if( Value ) {
      for( int i = 0; i < Collection->Count; i++ )
        if( Collection->Items[i]->Cancel )
          Collection->Items[i]->Cancel = false;
    }
    // set this button as the cancel button
    FCancel = Value;
  }
}
//---------------------------------------------------------------------------
void __fastcall TDcjButton::SetHelp(bool Value)
{
  // make sure it is not already set
  if( FHelp != Value ) {
    // if true, clear any other button set to the help button
    if( Value ) {
      for( int i = 0; i < Collection->Count; i++ )
        if( Collection->Items[i]->Help )
          Collection->Items[i]->Help = false;
    }
    // set this button as the help button
    FHelp = Value;
  }
}
//---------------------------------------------------------------------------
void __fastcall TDcjButton::SetModalResult(TModalResult Value)
{
  // make sure it is not already set
  if( FModalResult != Value )
    FModalResult = Value;
}
//---------------------------------------------------------------------------
void __fastcall TDcjButton::Assign(TPersistent *Source)
{
  // assign the properties from one TDcjButton to another
  if( Source->InheritsFrom(__classid(TDcjButton)) ) {
    // cast the TPersistant object to a TDcjButton object
    TDcjButton *source = dynamic_cast<TDcjButton *>(Source);
    // assign the TDcjButton properties
    ButtonID = source->ButtonID;
    Caption = source->Caption;
    Width = source->Width;
    Height = source->Height;
    Default = source->Default;
    Cancel = source->Cancel;
    Help = source->Help;
    ModalResult = source->ModalResult;
  }
}
//---------------------------------------------------------------------------
// TDcjMessageDlg Implementation
//---------------------------------------------------------------------------
__fastcall TDcjMessageDlg::TDcjMessageDlg(TComponent* Owner)
  : TComponent(Owner)
{
  FCaption = "";
  FDialogType = mtWarning;
  FFont = new TFont();
  FHelpContext = 0;
  // create the buttons container
  FButtonContainer = new TDcjButtonContainer(this);
}
//---------------------------------------------------------------------------
__fastcall TDcjMessageDlg::~TDcjMessageDlg(void)
{
  // free up the memory
  if(FFont) delete FFont;
  if(FButtonContainer) delete FButtonContainer;
}
//---------------------------------------------------------------------------
void __fastcall TDcjMessageDlg::Assign(TPersistent *Source)
{
  // assign the source object to the destination
  if( Source->InheritsFrom(__classid(TDcjMessageDlg)) ) {
    // cast the TPersistant object to a TDcjMessageDlg object
    TDcjMessageDlg *source = dynamic_cast<TDcjMessageDlg *>(Source);
    // assign the properties
    FCaption = source->Caption;
    FFont->Assign(source->Font);
    FMessage = source->Message;
    FDialogType = source->DialogType;
    FButtonContainer->Assign(source->ButtonContainer);
    FModalResult = source->ModalResult;
    FButtonID = source->ButtonID;
    FHelpContext = source->HelpContext;
  }
}
//---------------------------------------------------------------------------
bool __fastcall TDcjMessageDlg::Execute(void)
{
  // create the message dialog and return the result
  bool result = false;

  TDcjMessageForm *MsgDlg = CreateMessageDialog();
  try {
    // display the dialog and return the ButtonID and the modal result
    FModalResult = MsgDlg->ShowModal();
    FButtonID = MsgDlg->ButtonID;
    result = true;
  }
  catch(...) {
    delete MsgDlg;
    MsgDlg = NULL;
  }
  // free up the memory
  if(MsgDlg) {
    delete MsgDlg;
    MsgDlg = NULL;
  }

  return result;
}
//---------------------------------------------------------------------------
TDcjButtons* __fastcall TDcjMessageDlg::GetButtons(void)
{
  // return the ButtonContainer->Buttons object
  if( FButtonContainer )
    return  FButtonContainer->Buttons;
  else
    return NULL;
}
//---------------------------------------------------------------------------
TDcjMessageForm* __fastcall TDcjMessageDlg::CreateMessageDialog(void)
{
int const mcHorzMargin = 8;
int const mcVertMargin = 8;
int const mcHorzSpacing = 10;
int const mcVertSpacing = 10;
int iButtonWidth;
int iButtonHeight;
int iButtonMaxHeight = 0;

char *cpIcons[5] = {IDI_EXCLAMATION, IDI_HAND, IDI_ASTERISK,
    IDI_QUESTION, NULL};

  // creates the message dialog and retrieves the users selection
  TDcjMessageForm *MsgForm = new TDcjMessageForm(Application, 1);
  MsgForm->BorderStyle = bsDialog;
  MsgForm->Font->Assign(Font);
  MsgForm->HelpContext = HelpContext;

  // get the average character size
  TPoint DialogUnits;
  TSize du;
  char buffer[52];
  for( int i = 0; i < 26; i++ )
    buffer[i] = char(i + 'A');
  for( int i = 0; i < 26; i++ )
    buffer[i + 26] = char(i + 'a');
  GetTextExtentPoint32(MsgForm->Canvas->Handle, buffer, 52, &du);
  // because the stupid TSize and TPoint are incompatible under BCB, I
  // have to do this stupid assignment.
  DialogUnits.x = div(du.cx, 52).quot;
  DialogUnits.y = du.cy;

  int iHorzMargin = MulDiv(mcHorzMargin, DialogUnits.x, 4);
  int iVertMargin = MulDiv(mcVertMargin, DialogUnits.y, 8);
  int iHorzSpacing = MulDiv(mcHorzSpacing, DialogUnits.x, 4);
  int iVertSpacing = MulDiv(mcVertSpacing, DialogUnits.y, 8);
  int iButtonSpacing = MulDiv(ButtonContainer->Spacing, DialogUnits.x, 4);


  // set the height and width of the area needed to display the message
  // note: DrawText will not actually draw the text, because we are using
  // the DT_CALCRECT format flag.
  RECT TextRect;
  // set the area in which the text will be drawn.  maximum width is half of
  // the screen size, after that the message will word wrap.
  SetRect(&TextRect, 0, 0, div(Screen->Width, 2).quot, 0);
  DrawText(MsgForm->Canvas->Handle, Message.c_str(), -1, &TextRect,
      DT_CALCRECT | DT_WORDBREAK);

  // retrieve the icon and set the width and height needed for it
  char *cpIcon = cpIcons[DialogType];
  int iIconTextWidth = TextRect.right;
  int iIconTextHeight = TextRect.bottom;

  // check if the user is supplying a custom icon
  if( cpIcon != NULL ) {
    // set the width for the icon
    iIconTextWidth += 32 + iHorzSpacing;
    // make sure the icon height is at least 32 pixels
    if( iIconTextHeight < 32 )
      iIconTextHeight = 32;
  }

  // make a copy of the Buttons object.  we need to do this because we are
  // altering the buttons width and height settings by converting them to
  // dialog units.  we don't want to permanently alter the Buttons object or
  // the button sizes will grow every time the dialog is displayed.
  TDcjButtons *buttons = new TDcjButtons(this);
  buttons->Assign(Buttons);

  // get the total button width
  int iButtonCount = buttons->Count;
  int iButtonGroupWidth = 0;
  if( iButtonCount != 0 ) {
    // cycle through the buttons, retrieving the total width and max height
    for( int i = 0; i < iButtonCount; i++ ) {
      TDcjButton *button = buttons->Items[i];
      // convert from dialog units to pixels
      iButtonWidth = MulDiv(button->Width, DialogUnits.x, 4);
      iButtonHeight = MulDiv(button->Height, DialogUnits.y, 8);
      // store the dialog units in the button properties
      button->Width = iButtonWidth;
      button->Height = iButtonHeight;
      // total the button width
      iButtonGroupWidth += iButtonWidth;
      // get the max button height
      iButtonMaxHeight = max(iButtonMaxHeight, button->Height);
    }
    // include the spacing
    iButtonGroupWidth += iButtonSpacing * (iButtonCount - 1);
  }

  // set the form client height and width based on what is necessary to
  // display the buttons, text, and icon.
  MsgForm->ClientWidth = max(iIconTextWidth, iButtonGroupWidth) +
      iHorzMargin * 2;
  MsgForm->ClientHeight = iIconTextHeight + iButtonMaxHeight + iVertSpacing +
      iVertMargin * 2;

  // position the dialog in the center of the screen
  MsgForm->Left = (div(Screen->Width, 2).quot) - (div(MsgForm->Width, 2).quot);
  MsgForm->Top = (div(Screen->Height, 2).quot) - (div(MsgForm->Height, 2).quot);

  // set the dialog caption
  MsgForm->Caption = Caption;

  // create the icon image
  if( cpIcon != NULL ) {
    TImage *IconImage = new TImage(MsgForm);
    IconImage->Name = "imgIcon";
    IconImage->Parent = MsgForm;
    IconImage->Picture->Icon->Handle = LoadIcon(NULL, cpIcon);
    IconImage->SetBounds(iHorzMargin, iVertMargin, 32, 32);
  }

  // create the message label
  TLabel *MsgLabel = new TLabel(MsgForm);
  MsgLabel->Name = "Message";
  MsgLabel->Parent = MsgForm;
  MsgLabel->WordWrap = true;
  MsgLabel->Caption = Message;
  MsgLabel->BoundsRect = TextRect;
  MsgLabel->SetBounds(iIconTextWidth - TextRect.right + iHorzMargin,
      iVertMargin, TextRect.right, TextRect.bottom);

  // get the left margin needed to align the buttons
  int buttontop = 0;
  int buttonleft = 0;
  switch( ButtonContainer->Horizontal ) {
    case haLeft:
      buttonleft = iHorzMargin;
      break;
    case haCenter:
      buttonleft = div((MsgForm->ClientWidth - iButtonGroupWidth), 2).quot;
      break;
    case haRight:
      buttonleft = MsgForm->ClientWidth - (iButtonGroupWidth + iHorzMargin);
      break;
  }

  // create the buttons
  for( int i = 0; i < iButtonCount; i++ ) {
    // create a new button
    TButton *MsgBtn = new TButton(MsgForm);
    TDcjButton *button = buttons->Items[i];
    iButtonWidth = button->Width;
    iButtonHeight = button->Height;
    // set the button properties
    MsgBtn->Parent = MsgForm;
    MsgBtn->Caption = button->Caption;
    MsgBtn->ModalResult = button->ModalResult;
    MsgBtn->Default = button->Default;
    MsgBtn->Cancel = button->Cancel;
    // store the ButtonID in the tag property of the button.  This allows
    // me to set the ButtonID of the TDcjMessageDlg component when the
    // button is clicked.  This is not very elegant, but it works, and it
    // saves me from having to create my own TButton descendant.
    MsgBtn->Tag = button->ButtonID;
    if( button->Help )
      MsgBtn->OnClick = MsgForm->HelpButtonClick;
    else
      MsgBtn->OnClick = MsgForm->ButtonClick;
    // get the button top setting based on the vertical alignment
    switch( ButtonContainer->Vertical ) {
      case vaTop:
        buttontop = MsgForm->ClientHeight - iVertMargin - iButtonMaxHeight;
        break;
      case vaCenter:
        buttontop = MsgForm->ClientHeight - iVertMargin - iButtonMaxHeight +
            (div(iButtonMaxHeight - iButtonHeight, 2).quot);
        break;
      case vaBottom:
        buttontop = MsgForm->ClientHeight - iVertMargin - iButtonHeight;
        break;
    }
    // place and size the button
    MsgBtn->SetBounds(buttonleft, buttontop, iButtonWidth, iButtonHeight);
    // set the left margin for the next button
    buttonleft += iButtonWidth + iButtonSpacing;
  }
  // delete the copy of the Buttons object
  delete buttons;

  return MsgForm;
}
//---------------------------------------------------------------------------
void __fastcall TDcjMessageDlg::SetFont(TFont *Value)
{
  // set the font to the specified value
  FFont->Assign(Value);
}
//---------------------------------------------------------------------------
namespace Dcjmsgdg
{
  void __fastcall PACKAGE Register()
  {
    // register component
    TComponentClass classes[1] = {__classid(TDcjMessageDlg)};
    RegisterComponents("Dcj", classes, 0);
  }
}
//---------------------------------------------------------------------------

